/*
Continuous Testing Toolkit
Copyright (C) 2009 Paul R. Holser, Jr.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.infinitest.toolkit;

import org.junit.Before;
import org.junit.Test;
import static org.junit.Assert.*;

/**
 * Support structure for testing a class's compliance with the {@link Object#equals(Object) equals} and
 * {@link Object#hashCode() hashCode} contract.
 *
 * {@link #equal()} and {@link #notEqual()} methods should return objects that are of the same class.
 *
 * @author <a href="http://www.pholser.com">Paul Holser</a>
 */
public abstract class EqualsHashCodeTestSupport {
    private static final int NUMBER_OF_ITERATIONS = 20;

    protected Object firstEqual;
    protected Object secondEqual;
    protected Object thirdEqual;
    protected Object notEqual;

    /**
     * @return a new instance of the class under test; each object returned from this method should compare equal to
     *         each other, but not to the objects returned from {@link #notEqual()}.
     */
    protected abstract Object equal() throws Exception;

    /**
     * @return a new instance of the class under test; each object returned from this method should compare equal to
     *         each other, but not to the objects returned from {@link #equal()}.
     */
    protected abstract Object notEqual() throws Exception;

    @Before
    public final void initializeHarness() throws Exception {
        firstEqual = equal();
        secondEqual = equal();
        thirdEqual = equal();
        notEqual = notEqual();

        assertNotNull("equal() returned null", firstEqual);
        assertNotNull("2nd equal() returned null", secondEqual);
        assertNotNull("3rd equal() returned null", thirdEqual);
        assertNotNull("notEqual() returned null", notEqual);

        assertNotSame(firstEqual, secondEqual);
        assertNotSame(firstEqual, thirdEqual);
        assertNotSame(firstEqual, notEqual);
        assertNotSame(secondEqual, thirdEqual);
        assertNotSame(secondEqual, notEqual);
        assertNotSame(thirdEqual, notEqual);

        assertEquals("1st and 2nd equal instances of different classes", firstEqual.getClass(), secondEqual
            .getClass());
        assertEquals("1st and 3rd equal instances of different classes", firstEqual.getClass(), thirdEqual
            .getClass());
        assertEquals("1st equal instance and not-equal instance of different classes", firstEqual.getClass(),
            notEqual.getClass());
    }

    @Test
    public final void shouldNotBeEqualToJavaLangObject() {
        Object o = new Object();

        assertNotEquals(firstEqual, o);
        assertNotEquals(secondEqual, o);
        assertNotEquals(thirdEqual, o);
        assertNotEquals(notEqual, o);
    }

    @Test
    public final void shouldNotBeEqualToNull() {
        assertNotEquals("1st vs. null", firstEqual, null);
        assertNotEquals("2nd vs. null", secondEqual, null);
        assertNotEquals("3rd vs. null", thirdEqual, null);
        assertNotEquals("not-equal vs. null", notEqual, null);
    }

    @Test
    public final void unequalObjectsShouldNotCompareEqual() {
        assertNotEquals("1st vs. not-equal", firstEqual, notEqual);
        assertNotEquals("2nd vs. not-equal", secondEqual, notEqual);
        assertNotEquals("3rd vs. not-equal", thirdEqual, notEqual);

        assertNotEquals("not-equal vs. 1st", notEqual, firstEqual);
        assertNotEquals("not-equal vs. 2nd", notEqual, secondEqual);
        assertNotEquals("not-equal vs. 3rd", notEqual, thirdEqual);
    }

    @Test
    public final void equalsShouldBeConsistentAcrossInvocations() {
        for (int i = 0; i < NUMBER_OF_ITERATIONS; ++i) {
            shouldNotBeEqualToJavaLangObject();
            shouldNotBeEqualToNull();
            unequalObjectsShouldNotCompareEqual();
            equalsShouldBeReflexive();
            equalsShouldBeSymmetricAndTransitive();
        }
    }

    @Test
    public final void equalsShouldBeReflexive() {
        assertEquals("1st equal instance", firstEqual, firstEqual);
        assertEquals("2nd equal instance", secondEqual, secondEqual);
        assertEquals("3rd equal instance", thirdEqual, thirdEqual);
        assertEquals("not-equal instance", notEqual, notEqual);
    }

    @Test
    public final void equalsShouldBeSymmetricAndTransitive() {
        assertEquals("1st vs. 2nd", firstEqual, secondEqual);
        assertEquals("2nd vs. 1st", secondEqual, firstEqual);

        assertEquals("1st vs. 3rd", firstEqual, thirdEqual);
        assertEquals("3rd vs. 1st", thirdEqual, firstEqual);

        assertEquals("2nd vs. 3rd", secondEqual, thirdEqual);
        assertEquals("3rd vs. 2nd", thirdEqual, secondEqual);
    }

    @Test
    public final void equalObjectsShouldGiveSameHashCode() {
        assertEquals("1st vs. 2nd", firstEqual.hashCode(), secondEqual.hashCode());
        assertEquals("1st vs. 3rd", firstEqual.hashCode(), thirdEqual.hashCode());
        assertEquals("2nd vs. 3rd", secondEqual.hashCode(), thirdEqual.hashCode());
    }

    @Test
    public final void hashCodeShouldBeConsistentAcrossInvocations() {
        int firstHash = firstEqual.hashCode();
        int secondHash = secondEqual.hashCode();
        int thirdHash = thirdEqual.hashCode();
        int notEqualHash = notEqual.hashCode();

        for (int i = 0; i < NUMBER_OF_ITERATIONS; ++i) {
            assertEquals("1st equal instance", firstHash, firstEqual.hashCode());
            assertEquals("2nd equal instance", secondHash, secondEqual.hashCode());
            assertEquals("3rd equal instance", thirdHash, thirdEqual.hashCode());
            assertEquals("not-equal instance", notEqualHash, notEqual.hashCode());
        }
    }

    private static void assertNotEquals(Object expected, Object actual) {
        assertNotEquals(null, expected, actual);
    }

    protected static void assertNotEquals(String message, Object expected, Object actual) {
        if ((expected == null && actual == null) || (expected != null && expected.equals(actual)))
            failNotEquals(message, expected);
    }

    private static void failNotEquals(String message, Object expected) {
        String formatted = "";
        if (message != null)
            formatted = message + ' ';

        fail(formatted + "expected not equals to: <" + expected + '>');
    }
}
