/*
Continuous Testing Toolkit
Copyright (C) 2009 Paul R. Holser, Jr.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.infinitest.toolkit;

import java.util.Collection;
import java.util.Iterator;
import static java.util.Arrays.*;
import static java.util.Collections.*;

import org.junit.Before;
import org.junit.Test;
import static org.junit.Assert.*;

/**
 * Support structure for testing the {@linkplain java.util.Collections#unmodifiableCollection(Collection)
 * unmodifiability} of a collection. This is useful for verifying that accessor methods that return collections
 * do not allow direct adding or removing from the underlying collection field.
 *
 * @param <T> the type of the items in the collections involved
 * @author <a href="http://www.pholser.com">Paul Holser</a>
 */
public abstract class UnmodifiableCollectionTestSupport<T> {
    protected Collection<T> items;
    private T containedItem;

    @Before
    public final void initializeHarness() {
        this.items = newCollection();
        this.containedItem = containedItem();

        assertFalse("need a non-empty collection", items.isEmpty());
        assertTrue("need an item that is in the collection", items.contains(containedItem));
    }

    @Test(expected = UnsupportedOperationException.class)
    public final void shouldPreventAdd() {
        items.add(newItem());
    }

    @SuppressWarnings("unchecked")
    @Test(expected = UnsupportedOperationException.class)
    public final void shouldPreventAddAll() {
        items.addAll(asList(newItem()));
    }

    @Test(expected = UnsupportedOperationException.class)
    public final void shouldPreventClear() {
        items.clear();
    }

    @Test(expected = UnsupportedOperationException.class)
    public final void shouldPreventRemove() {
        items.remove(containedItem);
    }

    @SuppressWarnings("unchecked")
    @Test(expected = UnsupportedOperationException.class)
    public final void shouldPreventRemoveAll() {
        items.removeAll(asList(containedItem));
    }

    @Test(expected = UnsupportedOperationException.class)
    public final void shouldPreventRetainAll() {
        items.retainAll(emptyList());
    }

    @Test(expected = UnsupportedOperationException.class)
    public final void shouldPreventRemoveViaIterator() {
        Iterator<T> iter = items.iterator();
        iter.next();
        iter.remove();
    }

    /**
     * @return a collection to test for unmodifiability
     */
    protected abstract Collection<T> newCollection();

    /**
     * @return an item to attempt to add to the {@linkplain #newCollection() collection}
     */
    protected abstract T newItem();

    /**
     * @return an item that is in the {@linkplain #newCollection() collection}
     */
    protected abstract T containedItem();
}
